// JavaLife.java
// Tony Howat xargle@eh.org
//
// Implements the game of life as a java application
//
// (c) Tony Howat 1997
//

import java.awt.*;
import java.io.*;

// class   : lifecanvas
// purpose : deals with main life grid
//
class lifecanvas extends Canvas {
   javalife jl;                             // a reference to our parent
   public boolean lifeboard[][];            // the main life board
   public int boardxsize=5,boardysize=5;    // size of the board
   public int xblock,yblock;                // size of each cell on screen
   private boolean wrapx=false,wrapy=false; // wrapping method

   // constructor:
   // take a copy of our parent, we need this for referencing the images
   // we have loaded. use of jl is avoided as it isn't particularly good
   // practice.
   //
   public lifecanvas(javalife jlf)
   {
     jl=jlf;
   }

   // draw a bug in graphics area g at position x,y on the life grid.
   // if state is set then the bug is alive.
   //
   public void draw_bug(Graphics g, int x, int y, boolean state)
   {
     int xpos,ypos;
     int xgoff,ygoff;

     xpos=jl.xblock*x;
     ypos=jl.yblock*y;
     xgoff=(jl.xblock-jl.width)/2;  // workout the offset from xpos and..
     ygoff=(jl.yblock-jl.height)/2; // ypos to plot the bug in the centre

     g.setColor(Color.black);
     g.drawRect(xpos,
                ypos,
                jl.xblock,
                jl.yblock);
     g.drawImage(state ? jl.bugs[1] : jl.bugs[0],xgoff+xpos,ygoff+ypos,this);
   }

   // redraw the board
   //
   public void paint(Graphics g)
   {
     int x=0,y=0;
     for(x=0;x<jl.boardxsize;x++)
       for(y=0;y<jl.boardysize;y++)
         draw_bug(g,x,y,lifeboard[x][y]);
   }

   // we use mouseup events to identify where the user is clicking on our
   // grid.
   //
   public boolean mouseUp(Event e, int x, int y)
   {
     int xbug,ybug;

     xbug=x/xblock; // calculate co-ordinates in terms of our life grid.
     ybug=y/yblock;

     if(xbug<boardxsize && ybug<boardysize) {        // safety check
       lifeboard[xbug][ybug]=!lifeboard[xbug][ybug]; // invert state
       repaint();
     }

     return true;
   }

   // alter lifecanvas's local copy of the board size
   //
   public void boardsize(int x, int y)
   {
     boardxsize=x;
     boardysize=y;
     lifeboard=new boolean[boardxsize][boardysize];
   }

   // alter lifecanvas's local copy of the blocksize
   //
   public void blocksize(int x, int y)
   {
     xblock=x;
     yblock=y;
   }

   // alter lifecanvas's local copy of the wrap method
   //
   public void changewrap(boolean x, boolean y)
   {
     wrapx=x;
     wrapy=y;
   }

   // set a cell on on lifecanvas's grid (called when loading by javalife).
   //
   public void seton(int x, int y)
   {
     lifeboard[x][y]=true;
   }

   // clear the entire grid
   //
   public void cleargrid()
   {
     int x=0,y=0;
     for(x=0;x<boardxsize;x++)
       for(y=0;y<boardysize;y++)
         lifeboard[x][y]=false;
     repaint();
   }

   // lookup the state of a co-ordinate on the life grid. this function
   // takes account of any wrapping in use.
   //
   public boolean lookupstate(int x, int y)
   {
     int xx=x, yy=y;

     if(xx<0 && wrapx)
       xx=(boardxsize-1);
     else {
       if(xx<0)
         return false;
     }

     if(yy<0 && wrapy)
       yy=(boardysize-1);
     else {
       if(yy<0)
         return false;
     }

     if(y>(boardysize-1) && wrapy)
       yy=0;
     else {
       if(y>(boardysize-1))
         return false;
     }

     if(x>(boardxsize-1) && wrapx)
       xx=0;
     else {
       if(x>(boardxsize-1))
         return false;
     }

     // our two co-ords are obviously on the grid so return the state
     return(lifeboard[xx][yy]);
   }

   // calculate a generation
   //
   public void generate()
   {
     boolean nlifeboard[][]=new boolean[boardxsize][boardysize];     
     int x=0,y=0;

     for(x=0;x<boardxsize;x++)
     {
       for(y=0;y<boardysize;y++)
       {
         int neighbours=0;

         // find the number of neighbours for this cell.
         //
         if(lookupstate(x+1,y+1))
           neighbours++;
         if(lookupstate(x,y+1))
           neighbours++;
         if(lookupstate(x-1,y+1))
           neighbours++;
         if(lookupstate(x+1,y))
           neighbours++;
         if(lookupstate(x-1,y))
           neighbours++;
         if(lookupstate(x+1,y-1))
           neighbours++;
         if(lookupstate(x,y-1))
           neighbours++;
         if(lookupstate(x-1,y-1))
           neighbours++;
    
         /* compute whether this cell kicks the bucket or not */
      
         /* 0 or 1 neighbours and we die lonely */
         /* More than 3 neighbours and we die of overcrowding */
         if(neighbours<2 || neighbours>3) 
           nlifeboard[x][y]=false;
         else {
           /* if there are three neighbours and the current cell
            * isn't live a new cell is born - existing cells with
            * three or two neighbours survive */
           if(neighbours==3)
             nlifeboard[x][y]=true;
           else
             nlifeboard[x][y]=lifeboard[x][y];
         }  
       }
     }

     // copy new states to our main life table

     for(x=0;x<boardxsize;x++)
     {
       for(y=0;y<boardysize;y++)
       {
         lifeboard[x][y]=nlifeboard[x][y];         
       }
     }

     repaint();

     // we leave java to call the destructor for newlifetable
   }
}

// class   : javalife
// purpose : main class, deals with general UI and drives lifecanvas
// notes   : implemented as a frame based application rather than an
//           applet, allowing more sophisticated control over windows.

public class javalife extends Frame implements Runnable{
  public Image bugs[];                     // our images
  public int height,width;                 // width and height of images
  public int xblock,yblock;                // size of life cells
  public int boardxsize=10,boardysize=10;  // size of board in cells
  private MediaTracker imageTracker;       // for monitoring loading
  private lifecanvas c;                    // our main life display
  private Panel controlpanel;              // the control panel
  Button  generate,run,clear,load,save;    // buttons for control panel
  private TextField gridsize;              // grid size writable
  private Choice boardtype;                // board type choice list
  private TextField messages;              // field for messages
  Thread timer = null;                     // our timing thread
  boolean running=false;                   // is a sequence running?
  private FileDialog filed;                // our filedialog for load & save

  public javalife()
  {
    super("JavaLife");

    bugs=new Image[2];
  
    imageTracker=new MediaTracker(this);    
    setLayout(new BorderLayout());

    //tickthread=new tickerthread(1000,this);

    // set up choice component
    boardtype=new Choice();
    boardtype.addItem("Off board is dead");
    boardtype.addItem("Board wraps X + Y");
    boardtype.addItem("Board wraps X");
    boardtype.addItem("Board wraps Y");

    // set up control panel
    controlpanel=new Panel();
    generate=new Button("Generate");
    run=new Button("Run/Stop");
    clear=new Button("Clear");
    save=new Button("Save");
    load=new Button("Load");
    gridsize=new TextField(8);
    gridsize.setText("5,5");
    controlpanel.add(generate);
    controlpanel.add(run);
    controlpanel.add(clear);
    controlpanel.add(new Label("Grid Size (x,y)"));
    controlpanel.add(gridsize);
    controlpanel.add(boardtype);
    controlpanel.add(load);
    controlpanel.add(save);

    bugs[0]=Toolkit.getDefaultToolkit().getImage("dead.gif");
    bugs[1]=Toolkit.getDefaultToolkit().getImage("alive.gif");
    imageTracker.addImage(bugs[0],0);
    imageTracker.addImage(bugs[1],1);
    try {
      imageTracker.waitForID(0);
    } catch(InterruptedException e) {}
    try {
      imageTracker.waitForID(1);
    } catch(InterruptedException e) {}
 
    width=bugs[0].getWidth(this);
    height=bugs[0].getHeight(this);
    xblock=(int)((double)width*1.5);
    yblock=(int)((double)height*1.5);

    messages=new TextField(40);
    messages.setText("JavaLife, (c) Tony Howat 1997");
    messages.setEditable(false);

    c=new lifecanvas(this);
    c.boardsize(10,10);
    c.blocksize(xblock,yblock);
    c.cleargrid();
    c.setBackground(Color.white);
    c.resize(xblock*(boardxsize),yblock*(boardysize));    
    add("Center",c);
    add("South",controlpanel);
    add("North",messages);
   
    pack();
    setResizable(false);

    show();
    start();
   }

   // weeny function to put a message in the text field at the top of the
   // javalife window
   //
   public void message(String s)
   {
     messages.setText(s);
   }

   //////////////////////////////////////////////////////////////////////////
   // file loading and saving :
   //
   // could be much better. at the moment stores each life cell as an
   // integer, when they could be packed up as individual bits. unfortunately
   // i didn't have time to implement it.
   //
   // exceptions are carefully handled in both routines to avoid any
   // crashes.
   //////////////////////////////////////////////////////////////////////////

   // load a file.
   //
   public void file_load(String file)
   {
     DataInputStream input;
     int xsize,ysize,x,y;

     try {
       input = new DataInputStream(new FileInputStream(file));
     }
     catch (IOException e) {
       message("File not opened - "+e.toString());
       return;
     }

     try {
       // read the size of board we require and resize the board appropriately
       xsize=input.readInt();
       ysize=input.readInt();
       boardxsize=xsize;
       boardysize=ysize;
       c.boardsize(xsize,ysize);
       c.cleargrid();
       c.setBackground(Color.white);
       c.resize(xblock*(boardxsize),yblock*(boardysize));
       // make our frame as small as possible
       pack();

       // read in the individual cells
       for(x=0;x<xsize;x++)
       {
         for(y=0;y<ysize;y++)
         {
           if(input.readInt() > 0)
              c.seton(x,y);
         }
       }
     }
     catch (IOException e) {
       message("File error - "+e.toString());
       return;
     }     

     try {
       input.close();
     }
     catch (IOException e) {
       message("File not closed - "+e.toString());
       return;
     }

     message("File "+file+" loaded.");
   }

   // save a file
   //
   public void file_save(String file)
   {
     DataOutputStream output;
     int x,y;

     try {
       output = new DataOutputStream(new FileOutputStream(file));
     }
     catch (IOException e) {
       message("File not opened - "+e.toString());
       return;
     }

     try {
       output.writeInt(boardxsize); // write out the current board size
       output.writeInt(boardysize);

       // write out the cells
       for(x=0;x<boardxsize;x++)
       {
         for(y=0;y<boardysize;y++)
         {
           if(c.lookupstate(x,y))
             output.writeInt(1);
           else
             output.writeInt(0);
         }
       }
     }
     catch (IOException e) {
       message("File error - "+e.toString());
       return;
     }   

     try {
       output.flush(); // make sure nothing is left in any buffers
       output.close();
     }
     catch (IOException e) {
       message("File not closed - "+e.toString());
       return;
     }

     message("File "+file+" saved.");
   }

   // because this is a frame class we have to deal with window closing.
   // quite handy as we don't want to stick around when our main window
   // is closed.
   //
   public boolean handleEvent(Event e)
   {
     if(e.id==Event.WINDOW_DESTROY) {
       hide();
       dispose();
       System.exit(0);
       return true;
     }

     return super.handleEvent(e);
   }

   // action handler, identifies type and specific action and responds
   // accordingly
   //
   public boolean action(Event e, Object o)
   {
     if(e.target instanceof Button)
     {
       // grid clear button
       //
       if(e.arg.equals("Clear")) {
         c.cleargrid();
         message("Grid cleared");
       }

       // run/stop button
       //
       if(e.arg.equals("Run/Stop")) {
         if(running) {
           message("Stopped, click Run/Stop to start again");
         } else {
           message("Running at 1 generation per second. Click Run/Stop to stop");
         }
         running=!running; // (the thread code watches this boolean)
       }

       // the generate button
       //
       if(e.arg.equals("Generate")) {
         c.generate();
         message("Generated");
       }

       // the load button
       //
       if(e.arg.equals("Load")) {
         filed=new FileDialog(this,"Load Pattern",FileDialog.LOAD);
         filed.show(); // modal, so we are paused until we have a name
         file_load(filed.getDirectory()+filed.getFile()); 
       }

       // the save button
       //
       if(e.arg.equals("Save")) {
         filed=new FileDialog(this,"Save Pattern",FileDialog.SAVE);
         filed.show(); // modal, so we are paused until we have a name
         file_save(filed.getDirectory()+filed.getFile());
       }
     }

     // handle the choice component
     // (identify selection and alter x/y wrap as appropriate)
     //
     if(e.target instanceof Choice)
     {
       int index;

       index=boardtype.getSelectedIndex();
       switch(index)
       {
         case 0:
           message("Off board is now treated as dead");
           c.changewrap(false,false);
           break;
         case 1:
           message("Board now wraps in both directions");
           c.changewrap(true,true);
           break;
         case 2:
           message("Board now wraps horizontally");
           c.changewrap(true,false);
           break;
         case 3:
           message("Board now wraps vertically");
           c.changewrap(false,true);
           break;
       }
     }

     // a new grid size has been entered
     // parse it, validate, and set the new size
     //
     if(e.target == gridsize)
     {
       String entered=String.valueOf(o);
       String xstr,ystr;
       int commaindex, x, y;

       entered.trim();  // drop any whitespace
       // split the string
       commaindex=entered.indexOf(",");
       if(commaindex==-1)
       {
         message("Size should be in form (x,y)");
         return true;
       }
       xstr=entered.substring(0,commaindex);
       xstr=xstr.trim();  // drop any whitespace
       ystr=entered.substring(commaindex+1);
       ystr=ystr.trim();  // drop any whitespace

       try {
         y=Integer.parseInt(ystr);
         x=Integer.parseInt(xstr);
       }
       catch (NumberFormatException exception) {
         message("Size should be in form (x,y) "+xstr+" "+ystr);
         return true;
       }

       // range check
       if(x<2 || y<2 || x>100 || y>100) {
         message("x and y should be >2 and <100");
         return true;
       }

       // finally set the size and shrink/largen frame accordingly
       boardxsize=x;
       boardysize=y;
       c.boardsize(x,y);
       c.cleargrid();
       c.setBackground(Color.white);
       c.resize(xblock*(boardxsize),yblock*(boardysize));    
       pack();         
     }  
     return true;
   }

   // start our timer thread
   //
   public void start()
   {
     if(timer == null)
     {
       timer = new Thread(this);
       timer.start();
     }
   }

   // stop timer thread
   //
   public void stop()
   {
     timer = null;
   }

   // thread loop
   //
   public void run()
   {
     while (timer != null)
     {
       try {Thread.sleep(1000);} catch (InterruptedException e){}
       // if sequence is running go do the next generation
       //
       if(running)
         c.generate();
     }
     timer = null;
   }

   // main - construct our javalife instantation
   //
   public static void main(String args[])
   {
     javalife e;

     e=new javalife();
   }  
}
